{$X+}{$F+}{$D-}
unit Rootform;
interface

uses
  SysUtils, WinTypes, WinProcs, Toolhelp, Classes, Graphics, Controls,
  Forms, Consts, Typinfo, Menus;

type
  TRootForm = class(TForm)
  private
    FLoading : Boolean;
    FRegisteredClasses : TList;
    { When we load resources we will start at base class and load resources }
    { For each class.  FCurrentClass tells us which class we are on. }
    FCurrentClass : TClass;
    FCurrentLevel : Integer;
    FZonked : TComponent;
    FReader : TReader;
    FLoaded : TList;
    FMyName : String;
    Procedure LoadRes;
    Procedure ReadAState(Stream : TStream);
    Procedure ReadFormRes(ResName : String);
    Procedure AddLoaded(Component : TComponent);
    Procedure CallLoaded(Component : TComponent);
    procedure NameMagic(Reader: TReader; Component: TComponent; var Name: string); { This one is an event handler }
    Procedure AutoRegisterClasses;
    Procedure AutoDeregisterClasses;
    Procedure LoadDefaults(Component : TComponent);
  protected
    Procedure Loaded; override;
  public
    Constructor Create(AOwner : TComponent); override;
    Destructor Destroy; override;
    Property CurrentClass : TClass Read FCurrentClass;
    Property ClassLevel : Integer Read FCurrentLevel;
  end;

implementation

const
  Info='Intelligent Form Inheritance for Delphi 1.0'+#13+
       '(C) Sam Liddicott 1996 '#13'sam@csluk.demon.co.uk'#13'Demo Version';
  CopyRight='Ancestral Forms v1.11 Copyright Sam Liddicott 1996';

type
  PComponent=^TComponent;

  TPublicComponent=class(TPersistent)
  public
    FOwner: TComponent;
    FName: PString;
    FTag: Longint;
    FComponents: TList;
    FDesignInfo: Longint;
    FComponentState: TComponentState;
  public
    Property Owner: TComponent Read FOwner Write FOwner;
    Property Name: PString Read FName Write FName;
    Property Tag: Longint Read FTag Write FTag;
    Property Components: TList Read FComponents Write FComponents;
    Property DesignInfo: Longint Read FDesignInfo Write FDesignInfo;
    Property ComponentState: TComponentState Read FComponentState Write FComponentState;
  end;

  TPublicComponentChild=class(TComponent)
  public
    Procedure Loaded; override;
  end;

  TPublicReader = class(TFiler)
  public
    FOwner: TComponent;
    FParent: TComponent;
    FFixups: TList;
    FLoaded: TList;
  end;

  TPublicForm = class(TScrollingWinControl)
  private
    FActiveControl: TWinControl;
    FFocusedControl: TWinControl;
    FBorderIcons: TBorderIcons;
    FBorderStyle: TFormBorderStyle;
    FWindowState: TWindowState;
    FShowAction: TShowAction;
    FKeyPreview: Boolean;
    FActive: Boolean;
    FFormStyle: TFormStyle;
    FPosition: TPosition;
    FTileMode: TTileMode;
    FFormState: TFormState;
    FDropTarget: Boolean;
    FPrintScale: TPrintScale;
    FCanvas: TControlCanvas;
    FIcon: TIcon;
    FMenu: TMainMenu;
    FModalResult: TModalResult;
    FDesigner: TDesigner;
    FMenuHelp: THelpContext;
    FClientHandle: HWND;
    FWindowMenu: TMenuItem;
    FPixelsPerInch: Integer;
    FObjectMenuItem: TMenuItem;
    FHelper: TWinOleHelper;
    FClientWidth: Integer;
    FClientHeight: Integer;
    FTextHeight: Integer;
    FDefClientProc: TFarProc;
    FClientInstance: TFarProc;
    FOnActivate: TNotifyEvent;
    FOnClose: TCloseEvent;
    FOnCloseQuery: TCloseQueryEvent;
    FOnDeactivate: TNotifyEvent;
    FOnHide: TNotifyEvent;
    FOnPaint: TNotifyEvent;
    FOnResize: TNotifyEvent;
    FOnShow: TNotifyEvent;
    FOnCreate: TNotifyEvent;
    FOnDestroy: TNotifyEvent;
  end;

{ Class registration routines }
{ Nabbed from classes.pas }{ Who do they think they are, anyway? }
type
  PFieldClassTable = ^TFieldClassTable;
  TFieldClassTable = packed record
    Count: Smallint;
    Classes: array[0..8191] of TPersistentClass;
  end;

function GetFieldClassTable(AClass: TClass): PFieldClassTable; assembler;
asm
        XOR     AX,AX
        XOR     DX,DX
        LES     DI,AClass
        MOV     DI,ES:[DI-30]   { vtFieldTable }
        OR      DI,DI
        JE      @@1
        MOV     AX,ES:[DI+2]    { ftClassTable }
        MOV     DX,ES
@@1:
end;


Constructor TRootForm.Create;
begin
  CreateNew(AOwner);
  Include(TPublicForm(Self).FFormState, fsCreating);
  try
    { now loadres for all parents starting from root parent }
    LoadRes;
  finally
    Exclude(TPublicForm(Self).FFormState, fsCreating);
  end;
  try
    if Assigned(TPublicForm(Self).FOnCreate) then TPublicForm(Self).FOnCreate(Self);
  except
    Application.HandleException(Self);
  end;
  if fsVisible in TPublicForm(Self).FFormState then Visible := True;
end;

Procedure TRootForm.LoadRes;
var
  Parents : TList;
  C : Integer;
begin
  FLoading:=True;
  FLoaded:=TList.Create;
  AddLoaded(Self);
  FZonked:=TComponent.Create(nil);
  try
    FCurrentClass:=ClassType;
    Parents:=TList.Create;
    try
      repeat
        Parents.Add(FCurrentClass);
        FCurrentClass:=FCurrentClass.ClassParent;
      until (FCurrentClass=TRootForm);
      { Now loadres for list starting from end }
      FCurrentLevel:=Parents.Count-1;
      While FCurrentLevel>=0 do begin
        try
          FCurrentClass:=TClass(Parents[FCurrentLevel]);
          AutoRegisterClasses;
          try
            ReadFormRes(FCurrentClass.Classname);
          finally
            AutoDeregisterClasses;
          end;
        except
          On E: Exception do
            Application.ShowException(E);
        end;
        dec(FCurrentLevel);
      end;
      FLoading:=False;
      c:=0; while c<FLoaded.Count do begin
        CallLoaded(TComponent(FLoaded[c]));
        inc(c);
      end;
    finally
       Parents.Free;
    end;
  finally
    FLoaded.Free;
    FZonked.Free;
    FZonked:=nil;
  end;
end;

Destructor TRootForm.Destroy;
begin
  FZonked.Free;
  FRegisteredClasses.Free;
 Inherited Destroy;
end;

Procedure TRootForm.ReadFormRes;
  { We get a stream on the res }
var
  HResInfo: THandle;
  Handle: Integer;
  Stream: TStream;
  PCResName: array[0..63] of Char;
begin
  StrPLCopy(PCResName, ResName, SizeOf(PCResName) - 1);
  HResInfo := FindResource(HInstance, PCResName, rt_RCData);
  if HResInfo = 0 then
    raise EResNotFound.Create(FmtLoadStr(SResNotFound, [ResName]));
  Handle := AccessResource(HInstance, HResInfo);
  if Handle = 0 then
    raise EResNotFound.Create(FmtLoadStr(SResNotFound, [ResName]));
  try
    Stream := THandleStream.Create(Handle);
    try
      ReadAState(Stream);
    finally
      Stream.Free;
    end;
  finally
    FileClose(Handle);
  end;
end;

Procedure TRootForm.ReadAState;
begin
  { read this form from the stream! }
  FReader:=TReader.Create(Stream,1024);
  try
    FReader.OnSetName:=NameMagic;
    FReader.ReadRootComponent(Self);
  finally
    FReader.Free;
  end;
end;

Procedure TRootForm.Loaded;
var
  c : integer;
  LoadedList : TList;
begin
  { Fake outselves as loaded for every slot in FLoaded for TReader }
  If FCurrentLevel>=0 then begin
    LoadedList:=TPublicReader(FReader).FLoaded;
    c:=0; while c<LoadedList.Count do begin
      AddLoaded(LoadedList[c]);
      LoadedList[c]:=FZonked;
      inc(c);
    end;
    If FCurrentClass=Self.ClassType then
      Inherited Loaded;
  end;
end;

procedure TRootForm.AddLoaded;
begin
  { Don't add SELF to our loaded list as WE get called by the READER }
  If (FLoaded.IndexOf(Component)=-1) and (Component<>Self) then FLoaded.Add(Component);
end;

Procedure TRootForm.CallLoaded;
begin
  TPublicComponentChild(Component).Loaded;
end;

Procedure TRootForm.NameMagic;
var
  Ancestor : TComponent;
  Stack : ^Byte;
  StackP : Word;
  c : integer;
  Refs : TList;
begin
  try
    Component.Name:=Name;
  except
    { WHOA... MAGIC HERE }
    Ancestor:=FindComponent(Name);
    { Make sure they are of same class }
    If not (Component is Ancestor.ClassType) then begin
      Raise EComponentError.Create('Component in Descendant form '+Name+'('+
            Component.Classname+') incompatable with Ancestral '+Ancestor.Classname);
    end else begin
      { Step backwards through the stack to fint a pointer to Component }
      { Destroy it, and make it point to ancestor instead }
      asm
        mov StackP, BP
      end;
      Stack:=Ptr(SSeg,StackP);
      Refs:=TList.Create;
      try
        For c:=0 to 1000 do begin
          If pointer(PComponent(Stack)^)=pointer(Component) then begin
            If pointer(PComponent(Stack))=@Component then
            else begin
              Refs.Add(Stack);
            end;
          end;
          inc(Stack);
        end;
        { Now! Dispose of that rogue! }
        If Refs.Count>0 then begin
          Component.Free;
          c:=0; while c<refs.count do begin
            TComponent(Refs[c]^):=Ancestor;
            inc(c);
          end;
          { Fake csLoading - this is always done just after the naming }
          Include(TPublicComponent(Ancestor).FComponentState, csLoading);
          { Now fake csReading - a bit early, but we want to reload the }
          { defaults for all published properties }
          Include(TPublicComponent(Ancestor).FComponentState, csReading);
          LoadDefaults(Ancestor);
        end;
      finally
        Refs.Free;
      end;
    end; { of if right class }
  end;
end;

Procedure TRootForm.AutoRegisterClasses;
var
  I: Integer;
  ClassTable: PFieldClassTable;
  AClass : TPersistentClass;
begin
  FRegisteredClasses:=TList.Create;
  ClassTable := GetFieldClassTable(FCurrentClass);
  if ClassTable <> nil then
    for I:=0 to ClassTable^.Count - 1 do begin
      AClass:=ClassTable^.Classes[I];
      { If this class is not registered then register it and make a note }
      If GetClass(AClass.ClassName)=nil then begin
        FRegisteredClasses.Add(AClass);
        RegisterClass(AClass);
      end;
    end;
end;

Procedure TRootForm.AutoDeregisterClasses;
var
  c : Integer;
begin
  c:=0; while c<FRegisteredClasses.Count do begin
    UnRegisterClass(FRegisteredClasses[c]);
    inc(c);
  end;
  FRegisteredClasses.Free;
  FRegisteredClasses:=nil;
end;

Procedure TRootForm.LoadDefaults;
var
  I, Count: Integer;
  PropInfo: PPropInfo;
  PropList: PPropList;
begin
  Count := GetTypeData(Component.ClassInfo)^.PropCount;
  if Count > 0 then
  begin
    GetMem(PropList, Count * SizeOf(Pointer));
    try
      GetPropInfos(Component.ClassInfo, PropList);
      for I := 0 to Count - 1 do
      begin
        PropInfo := PropList^[I];
        if IsStoredProp(Component, PropInfo) then begin
          if PropInfo^.SetProc<>nil then begin
            { set default if ordinal }
            case PropInfo^.PropType^.Kind of
              tkInteger,tkChar,tkEnumeration,tkSet: begin
                SetOrdProp(Component, PropInfo, PropInfo^.Default);
              end;
            end
          end;
        end;
      end;
    finally
      FreeMem(PropList, Count * SizeOf(Pointer));
    end;
  end;
end;

procedure TPublicComponentChild.Loaded;
begin
  Inherited Loaded;
end;

Function GetParentPath : String;
Var
  taskEntry : TTaskEntry;
  Moduleentry : TModuleEntry;
begin
  TasKEntry.dwsize:=sizeof(taskentry);
  ModuleEntry.dwsize:=sizeof(moduleentry);
  TaskFindHandle(@TaskEntry,GetCurrentTask);
  TaskFindHandle(@TaskEntry,TaskEntry.htaskParent);
  ModuleFindHandle(@Moduleentry,TaskEntry.hModule);
  Result:=StrPas(ModuleEntry.szExePath);
end;

function X_MessageBox(WndParent: HWnd; Txt, Caption: PChar;
  TextType: Word): Integer; forward;
function X_MessageBox; external 'USER' index 1;

function U_MessageBox(WndParent: HWnd; Txt, Caption: String;
  TextType: Word): Integer;
var
  text : array[0..255] of char;
  title : array[0..255] of char;
begin
  Result:=X_MessageBox(WndParent,StrPCopy(title,txt),StrPCopy(text,caption),TextType);
end;

begin
  {$IFDEF DEMO}
  If not (ExtractFileName(GetParentPath)='DELPHI.EXE') then
    U_MessageBox(0,Info,CopyRight,MB_ICONINFORMATION or MB_OK);
  {$ENDIF}
end.


